SystemOrganization addCategory: #'Neo-Unicode-Collate'!

!NeoUnicodeDecomposingReadStream methodsFor: '*Neo-Unicode-Collate' stamp: 'HenrikSperreJohansen 1/28/2016 16:54'!
reset
	input reset.
	buffer removeAll.! !

!ZnLimitedReadStream methodsFor: '*Neo-Unicode-Collate' stamp: 'HenrikSperreJohansen 1/27/2016 12:53'!
peekFor: anElement
	^self peek = anElement and: [ self next. true ]! !

Object subclass: #CollateEntry
	instanceVariableNames: 'sequence combinations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!CollateEntry commentStamp: 'HenrikSperreJohansen 1/29/2016 19:32' prior: 0!
A CollateEntry holds both the sequence for a codepoint, and a map of the valid combinations with other characters that form different collate sequences/entries.

Users are responsible for resolving CollateEntries that are not to be further combined into CollateSequences, by calling #value.!

!CollateEntry methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/28/2016 17:03'!
at: aCodePoint ifAbsent: aBlock
	^combinations at: aCodePoint ifAbsent: aBlock! !

!CollateEntry methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/8/2016 15:44'!
at: aCodePoint put: anEntry
	combinations ifNil: [ combinations := SmallDictionary new ].
	combinations at: aCodePoint put: anEntry.! !

!CollateEntry methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/11/2016 10:29'!
followedBy: aCodePoint
	^combinations at: aCodePoint ifAbsent: [ nil ]! !

!CollateEntry methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/27/2016 13:13'!
followedBy: aCodePoint ifAbsentPut: anEntry 
	^(combinations ifNil: [combinations := Dictionary new]) at: aCodePoint ifAbsentPut: anEntry! !

!CollateEntry methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/14/2016 19:07'!
minimize
	"Return the minimal representation of me; if I have no combinations this is simply the sequence.
	If I have few combinations, use a linearly probed dictionary instead"

	combinations ifNil: [ ^ sequence ].
	combinations size <= 10
		ifTrue: [ combinations := SmallDictionary withAll: combinations ].
	combinations keysAndValuesDo: [ :key :value | combinations at: key put: value minimize ]! !

!CollateEntry methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/27/2016 13:14'!
setSequence: aSequence
	sequence := aSequence! !

!CollateEntry methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/28/2016 17:01'!
value
	^sequence! !

Object subclass: #CollateKeyStream
	instanceVariableNames: 'strategy nkdStream buffer level'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!CollateKeyStream commentStamp: 'HenrikSperreJohansen 1/11/2016 12:03' prior: 0!
CollateKeyStream provides a stream on the sort key  of a normalized string.
See http://unicode.org/reports/tr10/#Step_3.


!

!CollateKeyStream classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 1/27/2016 12:06'!
on: anNKDStream
	^self new initializeOn: anNKDStream! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 12:33'!
< anotherKeyStream
								
	^(self collateTo: anotherKeyStream) < 0! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 13:31'!
<= anotherKeyStream					
	^(self collateTo: anotherKeyStream) <= 0! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 12:33'!
= anotherKeyStream
								
	^(self collateTo: anotherKeyStream) = 0! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 12:33'!
> anotherKeyStream
								
	^(self collateTo: anotherKeyStream) > 0! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 12:33'!
>= anotherKeyStream
								
	^(self collateTo: anotherKeyStream) >= 0! !

!CollateKeyStream methodsFor: 'comparison' stamp: 'HenrikSperreJohansen 1/27/2016 12:31'!
collateTo: anotherKeyStream
	"return -1 if I am before anotherKeyStream, 0 if we are equal, and 1 if I am after"
	| a b |
	self strategy = anotherKeyStream strategy ifFalse: [ self error: 'Collating two sequences with different strategies is nonsensical' ].
	self reset.
	anotherKeyStream reset.
	[(a := self next) = (b := anotherKeyStream next) and: [b notNil]] whileTrue.
	^a ifNil:    [ b ifNil: [0] ifNotNil: [ 1] ]
		ifNotNil: [b ifNil: [-1] ifNotNil: [
				a < b ifTrue: [-1]
						ifFalse: [a = b ifTrue: [0]
											 ifFalse: [1]]]].! !

!CollateKeyStream methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/29/2016 20:06'!
fillBuffer
	"Read the next collate element at the current level from nkdStream into buffer"

	| entry candidates ccCheck isCC match |
	ccCheck := [ :codePoint | (NeoUnicodeCharacterData forCodePoint: codePoint) ifNotNil: [ :data | data ccc > 0 ] ifNil: [ false ] ].
	nkdStream next
		ifNil: [ buffer add: 0.
			level < strategy weight levels
				ifTrue: [ level := level + 1.
					nkdStream reset ] ]
		ifNotNil: [ :codePoint | entry := strategy collateEntryFor: codePoint ].	"Reached end of level"
	"Resolve combined entries. A combination is either a sequence of starters, or a sequence of starters ending in combining marks. Combining marks are special; the first one that has a combination will be used. See http://unicode.org/reports/tr10/#S2.1 "
	[ candidates := OrderedCollection with: nkdStream peek.
	isCC := ccCheck value: candidates last.
	[ isCC ]
		whileTrue:
			[ | candidate |
			candidate := nkdStream
				next;
				peek.
			isCC := ccCheck value: candidate.
			isCC
				ifTrue: [ candidates add: candidate ] ].
	match := candidates detect: [ :one | (entry followedBy: one) ifNotNil: [ :e | entry := e ] ] ifNone: nil.
	match ifNotNil: [ candidates remove: match ].
	match notNil and: [ candidates isEmpty ] ] whileFalse.

	"addLevel:in:precededBy:to: returns false if collate value is 0 for all elements at current level. Check if we any collate values are actually added from our entry + following cc's at the current level."
	"TODO: Check that no weigings does alterations on the boundaries we fill buffer, so nil can be used safely for first element "
	(candidates
		inject: (strategy weight addLevel: level in: entry precededBy: nil to: buffer)
		into: [ :sub :next | sub or: [ strategy weight addLevel: level in: (entry :=  strategy collateEntryFor: next) precededBy: entry to: buffer
				 ] ])
		ifFalse: [ self fillBuffer ]! !

!CollateKeyStream methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/27/2016 12:09'!
initialize
	strategy := CollateStrategy default.
	buffer := OrderedCollection new.
	level := 1.! !

!CollateKeyStream methodsFor: 'initialize-release' stamp: 'HenrikSperreJohansen 1/27/2016 13:35'!
initializeOn: anNKDStream

	nkdStream := anNKDStream.! !

!CollateKeyStream methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/8/2016 16:01'!
next
	buffer isEmpty ifTrue: [ self peek].
	^buffer isEmpty ifFalse: [ buffer removeFirst ]! !

!CollateKeyStream methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/11/2016 10:25'!
peek
	[buffer size < 2] whileTrue: [ self fillBuffer].  
	^buffer at: 2 ! !

!CollateKeyStream methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/29/2016 19:58'!
reset
	level := 1.

	nkdStream reset.
	buffer removeAll.! !

!CollateKeyStream methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/27/2016 12:09'!
strategy 
	^strategy! !

Object variableWordSubclass: #CollateSequence
	instanceVariableNames: ''
	classVariableNames: 'LevelMasks LevelShifts VariableWeightMask'
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!CollateSequence commentStamp: 'HenrikSperreJohansen 1/29/2016 19:29' prior: 0!
A CollateSequence is a sequence of Collate Elements.

Each collate element is stored as a word value, on the format: 
                                                
[level3][varWeigh][level2][level1]
[31,27 ][        26       ][ 25,16][ 15, 0  ].

If the DUCET or CLDR in newer versions go outside this range, adjustments to the class initialization will be necessary, hopefully 32bits will still be enough to sort the (strictly speaking, smaller) Unicode range.!

!CollateSequence classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 1/27/2016 13:11'!
fromElements: collectionOfElements
	^(self new: collectionOfElements size) initializeFrom: collectionOfElements! !

!CollateSequence classSide methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 1/27/2016 13:23'!
initialize
	"Each element in the DUCET has three levels.
	Use a word to store them all, masked and shifted as below.
	We use a bit to indicate whether the codepoint should be modified by variable weights."
	LevelShifts := #[0 16 26].
	LevelMasks := #(16rFFFF 16r1FF 16r3F).
	VariableWeightMask := 16r2000000! !

!CollateSequence methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/8/2016 15:51'!
at: anIndex level: aLevel
	^(self at: anIndex) >> (LevelShifts at: aLevel) bitAnd: (LevelMasks at: aLevel) ! !

!CollateSequence methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/27/2016 13:18'!
at: anIndex putLevel1: level1 level2: level2 level3: level3 variableWeight: aBoolean

	|collateElement|
	level1 > (LevelMasks at: 1) ifTrue: [ self error: 'Not enough room reserved for level1 collate values' ].
	level2 > (LevelMasks at: 2) ifTrue: [ self error: 'Not enough room reserved for level2 collate values' ].
	level3 > (LevelMasks at: 3) ifTrue: [ self error: 'Not enough room reserved for level3 collate values' ].
	collateElement := level1 << (LevelShifts at: 1).
	collateElement := collateElement + (level2 << (LevelShifts at: 2)).
	aBoolean ifTrue: [collateElement := collateElement bitAnd: VariableWeightMask].
	collateElement := collateElement + (level3 << (LevelShifts at: 3)).

	self at: anIndex put: collateElement
	! !

!CollateSequence methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 1/13/2016 14:30'!
initializeFrom: elementCollection
	"element = {level1. level2. level3. weight}"

	elementCollection
		withIndexDo:
			[ :element :index | 
			self
				at: index
				putLevel1: element first
				level2: element second
				level3: element third
				variableWeight: element last ]! !

!CollateSequence methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/11/2016 16:29'!
minimize
	"A sequence is already a minimal element"
	^self! !

Object subclass: #CollateStrategy
	instanceVariableNames: 'cetData variableWeight'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!CollateStrategy commentStamp: 'HenrikSperreJohansen 1/11/2016 12:08' prior: 0!
A SortStrategy defines  user-customizable parameters for how strings should be sorted.

At the moment, this includes:  

- Variable weight wrappers around returned collate sequences handling the defined cases http://unicode.org/reports/tr10/#Variable_Weighting!

!CollateStrategy classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:31'!
collateTable: aCollateTable
	"Use a strategy for collating that respects the given collate table"
	^self basicNew collateTable: aCollateTable! !

!CollateStrategy classSide methodsFor: 'defaults' stamp: 'HenrikSperreJohansen 12/14/2015 15:30'!
default
	^Default ifNil: [ Default :=  self collateTable: NeoUnicodeDUCETData default ]! !

!CollateStrategy classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:35'!
forLocale: aLocale
	^self collateTable: (NeoUnicodeDUCETData forLocale: aLocale)! !

!CollateStrategy classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:31'!
new
	^self collateTable: NeoUnicodeDUCETData default! !

!CollateStrategy methodsFor: 'weighing' stamp: 'HenrikSperreJohansen 1/27/2016 12:04'!
blanked
	variableWeight := Blanked! !

!CollateStrategy methodsFor: 'collating' stamp: 'HenrikSperreJohansen 1/4/2016 17:56'!
collate: aCollateStream to: anotherCollateStream
	"A very basic collate strategy"
	^(aCollateStream peek ) <= (anotherCollateStream peek )! !

!CollateStrategy methodsFor: 'collating' stamp: 'HenrikSperreJohansen 12/14/2015 13:15'!
collate: aCollateStream to: anotherCollateStream at: aCollateLevel
	"A very basic collate strategy"
	^(aCollateStream peek level: aCollateLevel) <= (anotherCollateStream peek level: aCollateLevel)! !

!CollateStrategy methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/28/2016 16:58'!
collateEntryFor: aCodePoint
	^cetData collateEntryFor: aCodePoint! !

!CollateStrategy methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:29'!
collateTable
	^cetData! !

!CollateStrategy methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 15:32'!
collateTable: aCollateTable
	cetData := aCollateTable! !

!CollateStrategy methodsFor: 'private  - sort delegation' stamp: 'HenrikSperreJohansen 12/14/2015 13:19'!
collectRanges: aCollateStreamCollection oldRanges: startRanges at: aLevel
	"Put this here, rather than in sorter, as I am uncertain whether collate strategies exist where ranges will have to be defined differently. 
	For now; the ranges are the elements after a sort that have the same collate value at the current level"
	|newRanges|
	newRanges := startRanges class new: startRanges size.
	startRanges do: [ :range | |rangeStart lastVal|
			rangeStart := range first.
			lastVal := (aCollateStreamCollection at: range first) peek level: aLevel.
			range first + 1 to: range last do: [:i | lastVal ~= (lastVal := (aCollateStreamCollection at: i) peek level: aLevel) ifTrue: [rangeStart - i > 0 ifTrue: [newRanges add: (rangeStart to: i)].
					rangeStart := i ]]].
		^newRanges! !

!CollateStrategy methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 1/27/2016 12:04'!
initialize
	variableWeight := NonIgnorable.! !

!CollateStrategy methodsFor: 'weighing' stamp: 'HenrikSperreJohansen 1/27/2016 12:04'!
nonIgnorable
	variableWeight := NonIgnorable! !

!CollateStrategy methodsFor: 'weighing' stamp: 'HenrikSperreJohansen 1/27/2016 12:04'!
shiftTrimmed
	variableWeight := ShiftTrimmed! !

!CollateStrategy methodsFor: 'weighing' stamp: 'HenrikSperreJohansen 1/27/2016 12:04'!
shifted
	variableWeight := Shifted! !

!CollateStrategy methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/29/2016 19:56'!
weight
	^variableWeight! !

Object subclass: #NeoUnicodeDUCETData
	instanceVariableNames: 'collateTable'
	classVariableNames: 'DUCET EmptyElement'
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!NeoUnicodeDUCETData commentStamp: 'HenrikSperreJohansen 1/15/2016 11:39' prior: 0!
This class holds the collate element data in the DUCET table (http://www.unicode.org/Public/UCA/latest/allkeys.txt). 
It also  creates derived collation elements when needed, see  http://unicode.org/reports/tr10/#S2.2

The table holds both CollateSequences, and CollateEntries, when loopups are fully resolved it is thus important to call #value in order to get the actual sequence in all cases.

Users should normally not consult the DUCET table data directly, but rather use a CollateStrategy, where one  can  both specify collate properties not contained in the collate element data, as well as potentially using different datatables  tailored for specific locales. 

!

!NeoUnicodeDUCETData classSide methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/15/2016 11:41'!
addCEFrom: allKeysStream to: collateTable
	|  codePoints element newSequence |
	"<entry>       := <charList> ';' <collElement>+ <eol>
	<collElement> := '[' <alt> <char> '.' <char> '.' <char> ('.' <char>)* ']'
	<alt>         := '*' | '.'"
	codePoints := OrderedCollection new.
	[allKeysStream peekFor: $;] 
		whileFalse: [ codePoints add: (Number readFrom: allKeysStream base: 16).
	"Skip next char if space"
	(allKeysStream peekFor: Character space)].
	(allKeysStream peekFor: Character space) ifFalse: [ self error: 'Strange format'].

	newSequence := self sequenceFrom: allKeysStream.
	element := collateTable.
	"Find the correct entry to add new sequence to."
	codePoints
		do:
			[ :codePoint | 
			element := element
				followedBy: codePoint
				ifAbsentPut: [ CollateEntry new ] ].
	element setSequence: newSequence.
	^ element! !

!NeoUnicodeDUCETData classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 11:24'!
default
	^DUCET ifNil: [self importDUCET ]! !

!NeoUnicodeDUCETData classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 15:36'!
forLocale: aLocale
	aLocale ifNil: [^self default].
	^NeoUnicodeLocaleCETData forLocale: aLocale! !

!NeoUnicodeDUCETData classSide methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/15/2016 11:40'!
importDUCET
	"http://unicode.org/reports/tr10/#File_Format
	<collationElementTable> := <version> 
                           <variable>?
                           <backwards>*
                           <entry>+
	@<version> := <major>.<minor>.<variant> <eol>
	<variable>       := '@variable ' <variableChoice> <eol>
	<variableChoice> := 'blanked' | 'non-ignorable' | 'shifted'
	<backwards> := ('@backwards ' | '@forwards ') <levelNumber> <eol>"

	| allKeysStream collateTable |
	allKeysStream := (ZnClient new
		streaming: true;
		get: 'http://www.unicode.org/Public/UCA/latest/allkeys.txt') ascii.
	"Skip properties preceding entries"
	[ '#@' includes: allKeysStream peek ]
		whileTrue: [ allKeysStream upTo: Character lf ].
	allKeysStream next.
	collateTable := CollateEntry new.
	[ allKeysStream atEnd ]
		whileFalse: [ self addCEFrom: allKeysStream to: collateTable ].
	collateTable minimize.
	^ DUCET := self new initializeFrom: collateTable! !

!NeoUnicodeDUCETData classSide methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 1/28/2016 02:50'!
initialize
	EmptyElement := CollateSequence fromElements: #((0 0 0 false))! !

!NeoUnicodeDUCETData classSide methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/27/2016 13:10'!
sequenceFrom: aCEStream
	"String on format '[*0032.0000.0000][.0000.0020.0000]'"
	| collateElements  |
	collateElements := OrderedCollection new.
	[aCEStream peekFor: $[] whileTrue: 
		[| weight level1 level2 level3| 
		weight := aCEStream next = $*.
		level1 := Number readFrom: (aCEStream next: 4) base: 16.
		aCEStream next = $. ifFalse: [ self error: 'Invalid file' ].
		level2 := Number readFrom: (aCEStream next: 4) base: 16.
		aCEStream next = $. ifFalse: [ self error: 'Invalid file' ].
		level3 := Number readFrom: (aCEStream next: 4) base: 16.
		aCEStream next = $] ifFalse: [ self error: 'Invalid file' ].
		collateElements add: { level1. level2. level3. weight }].
	"Skip to next line"
	aCEStream upTo: Character lf; next.
	^CollateSequence fromElements:  collateElements.
! !

!NeoUnicodeDUCETData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/11/2016 14:09'!
collateEntryFor: aCodePoint
	aCodePoint ifNil: [ ^ EmptyElement ].
	^ collateTable
		at: aCodePoint
		ifAbsent: [ self derivedCollateSequenceFor: aCodePoint ]! !

!NeoUnicodeDUCETData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/11/2016 14:07'!
derivedCollateSequenceFor: aCodePoint
	"Should create derived collate sequences, see http://unicode.org/reports/tr10/#S2.2"
	self notYetImplemented .
	^ EmptyElement! !

!NeoUnicodeDUCETData methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 1/27/2016 13:25'!
initializeFrom: aCollateTable
	collateTable := aCollateTable! !

Object subclass: #NeoUnicodeLocaleCETData
	instanceVariableNames: 'collateTable'
	classVariableNames: 'LocaleCETs'
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!NeoUnicodeLocaleCETData commentStamp: 'HenrikSperreJohansen 1/29/2016 19:38' prior: 0!
Instances of this class should hold Locale-specific alterations to the standard DUCET table.
The different locale modifications must be read from a suitable source, such as the CLDR base.

As the CLDR locale collate data format is a stateful list of modification to a root table, it is probably necessary with new classes (essentially a linked list with separate link classes representing level 1-4 differences) which is generated when parsing, and then converted into proper sequences to put in my instances. See http://unicode.org/reports/tr35/tr35-collation.html#Collation_Tailorings!

!NeoUnicodeLocaleCETData classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
createNewFor: aLocale
	^self new initializeFrom: (self loadCETDataFor: aLocale)! !

!NeoUnicodeLocaleCETData classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
forLocale: aLocale
	^LocaleCETs at: aLocale ifAbsentPut: [ self createNewFor: aLocale ]! !

!NeoUnicodeLocaleCETData classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 11:07'!
loadCETDataFor: aLocale
	"Find and parse the CLDR collate alterations for aLocale"
	^self notYetImplemented! !

!NeoUnicodeLocaleCETData methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/11/2016 14:09'!
collateEntryFor: aCharacter
	aCharacter
		ifNil: [ ^ NeoUnicodeDUCETData default collateEntryFor: aCharacter ].
	^ collateTable
		at: aCharacter codePoint
		ifAbsent: [ NeoUnicodeDUCETData default collateEntryFor: aCharacter ]! !

Object subclass: #NeoUnicodeSorter
	instanceVariableNames: 'strategy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!NeoUnicodeSorter commentStamp: 'HenrikSperreJohansen 1/29/2016 19:54' prior: 0!
NeoUnicodeSorter is an example implemention, meant to sorts strings according to a defined strategy.

The normal way to collate  would be by a Schwartzian transform; generating an association of CollateKeyStream -> string, sort these using methods in the CollateKeyStreams comparison category, then strio the CollateKeyStreams.
This will not require computation of full collate key arrays, but possibly recalculate the same collate sequence as a string is compared to multiple items.

This class instead sorts the input on the first collate key element of each string, then repeatedly re-sort by  subsequence collate key element the ranges that compared equal , until all ranges are sorted. 

TODO: Big-O comparison of worst-case  scenario for this approach vs the normal way (equally collating strings), and the importance  of using a search algorithm with good  performance in this case (O(n) vs O(n log n) for the viability of this case!

!NeoUnicodeSorter classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:34'!
forLocale: aLocale
	^self strategy: (CollateStrategy forLocale: aLocale)! !

!NeoUnicodeSorter classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 12:56'!
new
	^self strategy: CollateStrategy default! !

!NeoUnicodeSorter classSide methodsFor: 'instance creation' stamp: 'HenrikSperreJohansen 12/14/2015 15:27'!
strategy: aCollateStrategy
	^self basicNew initializeFor: aCollateStrategy! !

!NeoUnicodeSorter methodsFor: 'initialize - release' stamp: 'HenrikSperreJohansen 12/14/2015 15:28'!
initialize: aStrategy
	strategy := aStrategy! !

!NeoUnicodeSorter methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 1/8/2016 15:40'!
sort: aStringCollection
	| collateStreams |
	collateStreams := aStringCollection collect: [ :aString | CollateKeyStream on: aString collateTable: strategy collateTable ].
	self sort: collateStreams from: 1 to: collateStreams size.
	^collateStreams collect: #contents! !

!NeoUnicodeSorter methodsFor: 'private - sort' stamp: 'HenrikSperreJohansen 1/11/2016 13:51'!
sort: collateStreams from: start to: end
	| rangeStart element |
	collateStreams mergeSortFrom: start to: end by: [ :a :b | a peek < b peek ].
	rangeStart := start.
	element := (collateStreams at: rangeStart) next.
	start + 1 to: end do:
		[ :i | 
		element !!= (element := (collateStreams at: i) next)
			ifTrue:
				[ | rangeEnd |
				rangeEnd := i - 1.
				rangeEnd - rangeStart > 1
					ifTrue: [ self sort: collateStreams from: rangeStart to: rangeEnd.
						rangeStart := i ] ] ]! !

Object subclass: #VariableWeight
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!VariableWeight commentStamp: 'HenrikSperreJohansen 1/15/2016 11:05' prior: 0!
A variable weighing is a stateless, systematic modification to collate elements.

http://unicode.org/reports/tr10/#Variable_Weighting
!

VariableWeight subclass: #Blanked
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!Blanked commentStamp: 'HenrikSperreJohansen 1/15/2016 11:06' prior: 0!
Variable collation elements and any subsequent ignorable collation elements are reset so that all weights (except for the identical level) are zero. It is the same as the Shifted Option, except that there is no fourth level.!

!Blanked classSide methodsFor: 'constants' stamp: 'HenrikSperreJohansen 1/15/2016 11:23'!
levels
	^3 ! !

VariableWeight subclass: #NonIgnorable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!NonIgnorable commentStamp: 'HenrikSperreJohansen 1/15/2016 11:12' prior: 0!
Variable collation elements are not reset to be quaternary collation elements. All mappings defined in the table are unchanged.!

!NonIgnorable classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/15/2016 11:17'!
addLevel: aLevel in: aSequence precededBy: precedingSequence to: aBuffer
	"Return the amount of entries added"
	"Add all key elements to aBuffer, precedingBuffer unused"
	|total|
	total := 0.
	1 to: aSequence size do: [:i | |key|
		key := aSequence at: i level: aLevel.
		key > 0 ifTrue: [aBuffer add: key.
			 total := total + 1]].
	^total! !

!NonIgnorable classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/15/2016 11:15'!
collateKeyOf: sequence at: index level: aLevel precededBy: precedingSequence
	^sequence value at: index level: aLevel! !

!NonIgnorable classSide methodsFor: 'constants' stamp: 'HenrikSperreJohansen 1/15/2016 11:23'!
levels
	^3! !

VariableWeight subclass: #Shifted
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!Shifted commentStamp: 'HenrikSperreJohansen 1/15/2016 11:11' prior: 0!
Variable collation elements are reset to zero at levels one through three. In addition, a new fourth-level weight is appended, whose value depends on the type, as shown in Table 11. Any subsequent primary or secondary ignorables following a variable are reset so that their weights at levels one through four are zero.

Table 11. L4 Weights for Shifted Variables:
Type								L4		Examples
L1, L2, L3 = 0							0000	null [.0000.0000.0000.0000]
L1 ≠ 0, Variable						old L1	space [.0000.0000.0000.0209]
L1=0, L3 ≠ 0, following a Variable		0000	combining grave [.0000.0000.0000.0000]
L1 = 0, L3 ≠ 0, not following a Variable	FFFF	combining grave [.0000.0035.0002.FFFF]
L1 ≠ 0, not Variable					FFFF	Capital A [.06D9.0020.0008.FFFF]!

Shifted subclass: #ShiftTrimmed
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Collate'!

!ShiftTrimmed commentStamp: 'HenrikSperreJohansen 1/15/2016 11:07' prior: 0!
This option is the same as Shifted, except that all trailing FFFFs are trimmed from the sort key. This could be used to emulate POSIX behavior, but is otherwise not recommended.!

!ShiftTrimmed classSide methodsFor: 'constants' stamp: 'HenrikSperreJohansen 1/15/2016 11:24'!
levels
	^4! !

!Shifted classSide methodsFor: 'constants' stamp: 'HenrikSperreJohansen 1/15/2016 11:24'!
levels
	^4! !

!VariableWeight classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/15/2016 11:14'!
addLevel: aLevel in: aSequence precededBy: precedingSequence to: aStream
	"Add all entries in aSequence at aLevel to aStream, respecting my weighting.
	Return the amount of entries added"
	^self subclassResponsibility! !

!VariableWeight classSide methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 1/15/2016 11:14'!
collateKeyOf: sequence at: index level: aLevel precededBy: precedingSequence
	^self subclassResponsibility! !

!VariableWeight classSide methodsFor: 'constants' stamp: 'HenrikSperreJohansen 1/15/2016 11:23'!
levels
	"How many levels does a collate element support using this variable weight scheme?"
	^self subclassResponsibility! !
CollateSequence initialize!
NeoUnicodeDUCETData initialize!
